import merge from 'lodash/merge';
import mapKeys from 'lodash/mapKeys';
import each from 'lodash/each';
import pickBy from 'lodash/pickBy';
import omit from 'lodash/omit';
import isEqual from 'lodash/isEqual';
import { CssSelectorParser } from 'css-selector-parser';
import dedent from 'dedent';

const parser = new CssSelectorParser();
parser.registerSelectorPseudos('has');
parser.registerNestingOperators('>', '+', '~');
parser.registerAttrEqualityMods('^', '$', '*', '~');
parser.enableSubstitutes();
const validateSelector = selector => selector.split(',').map(selectorPart => {
  if (selectorPart.indexOf('&') === -1) {
    throw new Error(dedent`
          Invalid selector: ${selector}
      
          Selectors must target the ampersand character ('&'), which refers to the generated class name, e.g. '&:nth-child(2n)'
        `);
  }

  const ampersand = '__treat_ampersand__';
  let selectorAst;

  try {
    selectorAst = parser.parse(selectorPart.replace(/&/g, `.${ampersand}`));
  } catch (err) {
    throw new Error(`Invalid selector: ${selector}`);
  }

  let targetRule = selectorAst;

  while (targetRule.rule) {
    targetRule = targetRule.rule;
  }

  if (!Array.isArray(targetRule.classNames) || !targetRule.classNames.find(className => className === ampersand)) {
    throw new Error(dedent`
          Invalid selector: ${selector}
      
          Style selectors must end with the '&' character (along with any modifiers), e.g. ${'`${parent} &`'} or ${'`${parent} &:hover`'}.
          
          This is to ensure that each style block only affects the styling of a single class.
          
          If your selector is targeting another class, you should move it to the style definition for that class, e.g. given we have styles for 'parent' and 'child' elements, instead of adding a selector of ${'`& ${child}`'}) to 'parent', you should add ${'`${parent} &`'} to 'child').
          
          If your selector is targeting something global, use the 'globalStyle' function instead, e.g. if you wanted to write ${'`& h1`'}, you should instead write 'globalStyle(${'`${parent} h1`'}, { ... })'
        `);
  }
});

const simplePseudos = [':-moz-any-link', ':-moz-full-screen', ':-moz-placeholder', ':-moz-read-only', ':-moz-read-write', ':-ms-fullscreen', ':-ms-input-placeholder', ':-webkit-any-link', ':-webkit-full-screen', '::-moz-placeholder', '::-moz-progress-bar', '::-moz-range-progress', '::-moz-range-thumb', '::-moz-range-track', '::-moz-selection', '::-ms-backdrop', '::-ms-browse', '::-ms-check', '::-ms-clear', '::-ms-fill', '::-ms-fill-lower', '::-ms-fill-upper', '::-ms-reveal', '::-ms-thumb', '::-ms-ticks-after', '::-ms-ticks-before', '::-ms-tooltip', '::-ms-track', '::-ms-value', '::-webkit-backdrop', '::-webkit-input-placeholder', '::-webkit-progress-bar', '::-webkit-progress-inner-value', '::-webkit-progress-value', '::-webkit-slider-runnable-track', '::-webkit-slider-thumb', '::after', '::backdrop', '::before', '::cue', '::first-letter', '::first-line', '::grammar-error', '::placeholder', '::selection', '::spelling-error', ':active', ':after', ':any-link', ':before', ':blank', ':checked', ':default', ':defined', ':disabled', ':empty', ':enabled', ':first', ':first-child', ':first-letter', ':first-line', ':first-of-type', ':focus', ':focus-visible', ':focus-within', ':fullscreen', ':hover', ':in-range', ':indeterminate', ':invalid', ':last-child', ':last-of-type', ':left', ':link', ':only-child', ':only-of-type', ':optional', ':out-of-range', ':placeholder-shown', ':read-only', ':read-write', ':required', ':right', ':root', ':scope', ':target', ':valid', ':visited'];
const simplePseudoSet = new Set(simplePseudos);

const normalizeStyles = (className, styles) => {
  const omitThese = [...simplePseudos, '@media', '@supports', 'selectors'];
  const pseudoStyles = mapKeys(pickBy(styles, (_, key) => simplePseudoSet.has(key)), (_, pseudo) => `${className}${pseudo}`);
  let selectorStyles = {};

  if (styles.selectors) {
    selectorStyles = mapKeys(styles.selectors, (_, selector) => {
      // Themed selectors can be empty if themes haven't registered yet.
      // In this case don't validate the selector
      if (selector.length > 0) {
        validateSelector(selector);
      }

      return selector.replace(RegExp('&', 'g'), className);
    });
  }

  const rawRules = omit(styles, omitThese);
  const rawStyles = Object.keys(rawRules).length > 0 ? {
    [className]: rawRules
  } : {};
  const allStyles = Object.assign(rawStyles, pseudoStyles, selectorStyles);
  Object.keys(allStyles).forEach(ident => {
    if (allStyles[ident]['@keyframes']) {
      const {
        '@keyframes': keyframeRef,
        animation
      } = allStyles[ident];

      if (keyframeRef) {
        Object.assign(allStyles[ident], {
          animation: animation ? animation.replace('@keyframes', keyframeRef) : undefined,
          animationName: animation ? undefined : keyframeRef,
          '@keyframes': undefined
        });
      }
    }
  });
  return allStyles;
};

var transformCSS = (styles => {
  const stylesheet = {};
  const responsiveStylesheet = {};
  Object.entries(styles).forEach(([className, styles]) => {
    const defaultStyles = normalizeStyles(className, styles);
    const responsiveStyles = {};

    if (styles['@media']) {
      each(styles['@media'], (mediaStyles, query) => {
        const blockStyles = normalizeStyles(className, mediaStyles);

        if (!isEqual(defaultStyles, blockStyles)) {
          merge(responsiveStyles, {
            [`@media ${query}`]: blockStyles
          });
        }
      });
    }

    if (styles['@supports']) {
      each(styles['@supports'], (mediaStyles, query) => {
        const blockStyles = normalizeStyles(className, mediaStyles);

        if (!isEqual(defaultStyles, blockStyles)) {
          merge(responsiveStyles, {
            [`@supports ${query}`]: blockStyles
          });
        }
      });
    }

    merge(stylesheet, defaultStyles);
    merge(responsiveStylesheet, responsiveStyles);
  });
  return Object.assign(stylesheet, responsiveStylesheet);
});

export { simplePseudos as s, transformCSS as t };
