import { fromPairs, isEqual } from 'lodash';
import dedent from 'dedent';
import { stringify } from 'javascript-stringify';
import { getIdentName, getNextScope, addLocalCss, getThemes, addTheme, addThemedCss } from '../adapter/dist/treat-adapter.esm.js';
import { c as createContentHash } from './createContentHash-44e0456c.esm.js';
import mapKeys from 'lodash/mapKeys';
import flatMap from 'lodash/flatMap';
import uniq from 'lodash/uniq';
import joi from '@hapi/joi';
import { s as simplePseudos } from './transformCSS-97068520.esm.js';
import 'object-hash';
import 'lodash/merge';
import 'lodash/each';
import 'lodash/pickBy';
import 'lodash/omit';
import 'lodash/isEqual';
import 'css-selector-parser';

const themePlaceholder = '$';
const templateThemeClassRef = classRef => `${themePlaceholder}${classRef}`;
const makeThemedClassReference = (themeRef, classRef) => `${classRef}${themeRef}`;
const convertToCssClass = ref => `.${ref}`;

const resolveClassName = (themeRef, classRef) => classRef[0] === '$' ? makeThemedClassReference(themeRef, classRef.substring(1)) : classRef;

const cache = new Map();
const resolveStyles = (themeRef, styles) => {
  let themeCache = cache.get(themeRef);

  if (themeCache) {
    const cachedResolvedStyles = themeCache.get(styles);

    if (cachedResolvedStyles) {
      return cachedResolvedStyles;
    }
  } else {
    themeCache = new Map();
    cache.set(themeRef, themeCache);
  }

  const resolvedStyles = walkTreatModule(themeRef, styles);
  themeCache.set(styles, resolvedStyles);
  return resolvedStyles;
};

const walkTreatModule = (themeRef, styles) => {
  const clone = styles.constructor();

  for (let key in styles) {
    const value = styles[key];

    if (typeof value === 'string') {
      clone[key] = resolveClassName(themeRef, value);
    } else if (typeof value === 'object') {
      clone[key] = value ? walkTreatModule(themeRef, value) : value;
    } else {
      clone[key] = value;
    }
  }

  return clone;
};

const localClassRefs = [];
const isThemedSelector = selector => selector.indexOf(themePlaceholder) > -1;
const addLocalClassRef = classRef => localClassRefs.push(classRef);
const interpolateSelector = (selector, themeRef) => {
  let normalisedSelector = selector;

  if (localClassRefs.length > 0) {
    const localClassRefsRegex = RegExp(`(${localClassRefs.join('|')})`, 'g');
    normalisedSelector = normalisedSelector.replace(localClassRefsRegex, (_, match) => {
      return convertToCssClass(match);
    });
  }

  const themeClassRefsRegex = RegExp(`\\${themePlaceholder}([a-zA-Z0-9_-]+)`, 'g');
  return normalisedSelector.replace(themeClassRefsRegex, (_, match) => {
    if (!themeRef) {
      throw new Error(`No theme ref provided to 'interpolateSelector'`);
    }

    return convertToCssClass(makeThemedClassReference(themeRef, match));
  });
};
const combinedThemeSelector = (selector, themes) => {
  if (isThemedSelector(selector)) {
    return uniq(flatMap(selector.split(','), selectorPart => themes.map(({
      themeRef
    }) => interpolateSelector(selectorPart.trim(), themeRef)))).join(', ');
  }

  return interpolateSelector(selector);
};
const processSelectors = ({
  style,
  themeRef,
  themes
}) => {
  if (style.selectors) {
    style.selectors = mapKeys(style.selectors, (_valid, key) => themeRef ? interpolateSelector(key, themeRef) : combinedThemeSelector(key, themes));
  }

  const media = style['@media'];

  if (media) {
    Object.keys(media).forEach(mediaQuery => {
      processSelectors({
        style: media[mediaQuery],
        themeRef,
        themes
      });
    });
  }
};

const cssPropertiesSchema = joi.object().pattern(joi.string().regex(/^[^:@]/), [joi.string(), joi.number()]);
const keyframesSchema = joi.object().pattern(joi.string(), cssPropertiesSchema);
const cssWithKeyframesSchema = cssPropertiesSchema.append({
  '@keyframes': [joi.string(), keyframesSchema]
});
const cssWithSimplePseudos = cssWithKeyframesSchema.append(Object.assign({}, ...simplePseudos.map(pseudo => ({
  [pseudo]: cssWithKeyframesSchema
}))));
const selectorsSchema = joi.object().pattern(joi.string(), cssWithKeyframesSchema);
const fullStyle = cssWithSimplePseudos.append({
  selectors: selectorsSchema
});

const makeQuerySchema = valueSchema => valueSchema.append({
  '@media': joi.object().pattern(joi.string(), valueSchema)
}).append({
  '@supports': joi.object().pattern(joi.string(), valueSchema)
});

const fullStyleWithMedia = makeQuerySchema(fullStyle).required();
const globalStyleWithMedia = makeQuerySchema(fullStyle.fork(['selectors', ...simplePseudos], a => a.forbidden())).required();

const validate = handler => value => {
  const {
    error
  } = handler.validate(value);

  if (error) {
    throw error;
  }
};

const validateGlobalStyle = validate(globalStyleWithMedia);
const validateStyle = validate(fullStyleWithMedia);

const createKeyframe = (keyframe, themeRef) => {
  if (typeof keyframe === 'string') {
    return keyframe;
  }

  const keyframeRef = createContentHash(keyframe);
  const keyframeBlock = {
    [`@keyframes ${keyframeRef}`]: keyframe
  };

  if (themeRef) {
    addThemedCss(themeRef, keyframeBlock);
  } else {
    addLocalCss(keyframeBlock);
  }

  return keyframeRef;
};

const processAnimations = (style, themeRef) => {
  // handle root level keyframes
  if (style['@keyframes']) {
    style['@keyframes'] = createKeyframe(style['@keyframes'], themeRef);
  }

  const media = style['@media']; // handle keyframes in media queries

  if (media) {
    Object.keys(media).forEach(mediaQuery => {
      processAnimations(media[mediaQuery]);
    });
  } // handle keyframes in simple pseudos


  Object.entries(style).filter(([property]) => property.startsWith(':')).forEach(([_pseudoProperty, pseudoStyle]) => {
    if (pseudoStyle['@keyframes']) {
      pseudoStyle['@keyframes'] = createKeyframe(pseudoStyle['@keyframes'], themeRef);
    }
  }); // handle keyframes in complex selectors

  if (style.selectors) {
    Object.values(style.selectors).forEach(selectorStyle => {
      if (selectorStyle['@keyframes']) {
        selectorStyle['@keyframes'] = createKeyframe(selectorStyle['@keyframes'], themeRef);
      }
    });
  }
};

const processStyle = (style, themeRef) => {
  processSelectors({
    style,
    themeRef,
    themes: getThemes()
  });
  processAnimations(style, themeRef);
};

const createThemedCss = (classRef, style) => {
  Object.entries(style).forEach(([themeRef, themeStyle]) => {
    validateStyle(themeStyle);
    processStyle(themeStyle, themeRef);
    const themedClassRef = makeThemedClassReference(themeRef, classRef);
    addThemedCss(themeRef, {
      [convertToCssClass(themedClassRef)]: themeStyle
    });
  });
  return templateThemeClassRef(classRef);
};

function style(style, localDebugName) {
  const localName = localDebugName || 'style';
  const classRef = getIdentName(localName, getNextScope());

  if (typeof style === 'object') {
    validateStyle(style);
    addLocalClassRef(classRef);
    processStyle(style);
    addLocalCss({
      [convertToCssClass(classRef)]: style
    });
    return classRef;
  } else {
    const themedStyle = Object.assign({}, ...getThemes().map(({
      themeRef,
      tokens
    }) => {
      const themedStyle = style(tokens);
      validateStyle(themedStyle);
      return {
        [themeRef]: themedStyle
      };
    }));
    return createThemedCss(classRef, themedStyle);
  }
}
function styleMap(stylesheet, localDebugName) {
  const classRefs = {};

  const createLocalName = classIdentifier => {
    if (localDebugName) {
      return `${localDebugName}_${classIdentifier}`;
    }

    return classIdentifier;
  };

  if (typeof stylesheet === 'function') {
    const styleMap = new Map();
    getThemes().forEach(({
      themeRef,
      tokens
    }) => {
      const themedStyles = stylesheet(tokens);
      Object.entries(themedStyles).forEach(([classIdent, style]) => {
        const styleMapValue = styleMap.get(classIdent) || {};
        styleMap.set(classIdent, Object.assign({}, styleMapValue, {
          [themeRef]: style
        }));
      });
    });
    Array.from(styleMap.entries()).forEach(([classIdent, styles]) => {
      const classRef = getIdentName(createLocalName(classIdent), getNextScope());
      classRefs[classIdent] = createThemedCss(classRef, styles);
    });
  } else {
    const postCss = fromPairs(Object.entries(stylesheet).map( // @ts-expect-error
    ([classIdentifier, style]) => {
      validateStyle(style);
      const classRef = getIdentName(createLocalName(classIdentifier), getNextScope());
      classRefs[classIdentifier] = classRef;
      addLocalClassRef(classRef);
      processStyle(style);
      return [convertToCssClass(classRef), style];
    }));
    addLocalCss(postCss);
  } // @ts-expect-error


  return classRefs;
}
function createTheme(tokens, localDebugName) {
  const theme = {
    themeRef: getIdentName(localDebugName || 'theme', getNextScope(), tokens),
    tokens
  };
  addTheme(theme);
  return theme.themeRef;
}
function globalStyle(selector, style) {
  if (isThemedSelector(selector)) {
    getThemes().forEach(theme => {
      const themedSelector = interpolateSelector(selector, theme.themeRef);
      const themeStyle = typeof style === 'function' ? style(theme.tokens) : style;
      validateGlobalStyle(themeStyle);
      addThemedCss(theme.themeRef, {
        [themedSelector]: themeStyle
      });
    });
  } else {
    const normalisedSelector = interpolateSelector(selector);

    if (typeof style === 'function') {
      throw Error(dedent`
          Unthemeable selector: ${normalisedSelector}
          
          Global styles cannot be themed unless the selector references a themed class.
          
          For example, you could add a themed class to the start of your selector (e.g. ${'`${themedClass} h1`'}).
          
          Without a reference to a themed class, there is no way to selectively apply styles based on the current theme.
        `);
    }

    validateGlobalStyle(style);
    processAnimations(style);
    addLocalCss({
      [normalisedSelector]: style
    });
  }
}
function styleTree(makeStyleTree) {
  const themedClassRefs = new Map();
  const assignedScopes = [];

  const getNodeScope = scopeIndex => {
    if (typeof assignedScopes[scopeIndex] !== 'number') {
      assignedScopes[scopeIndex] = getNextScope();
    }

    return assignedScopes[scopeIndex];
  };

  const themedTrees = getThemes().map(({
    tokens,
    themeRef
  }) => {
    let scopeIndex = 0;

    const makeStyle = (style, localDebugName) => {
      const localName = localDebugName || 'styleNode';
      const classRef = getIdentName(localName, getNodeScope(scopeIndex++));
      const themedClassRefValue = themedClassRefs.get(classRef) || {};
      validateStyle(style);
      themedClassRefs.set(classRef, Object.assign({}, themedClassRefValue, {
        [themeRef]: style
      }));
      return templateThemeClassRef(classRef);
    };

    return makeStyleTree(tokens, makeStyle);
  });
  const [referenceTree, ...restTrees] = themedTrees;
  restTrees.forEach(tree => {
    if (!isEqual(tree, referenceTree)) {
      throw new Error(dedent`
        Mismatching style trees.

        All 'styleTree' functions must return the same structure for every theme.
        
        To avoid this error, ensure that object keys and array lengths do not depend on unique properties of each theme.
        
        Expected:
        
        ${stringify(referenceTree, null, 2)}

        Received:
        
        ${stringify(tree, null, 2)}
      `);
    }
  });
  Array.from(themedClassRefs.entries()).forEach(([classRef, styles]) => {
    createThemedCss(classRef, styles);
  });
  return referenceTree;
}

const createTheme$1 = createTheme;
const style$1 = style;
const styleMap$1 = styleMap;
const styleTree$1 = styleTree;
const globalStyle$1 = globalStyle;

export { createTheme$1 as createTheme, globalStyle$1 as globalStyle, resolveClassName, resolveStyles, style$1 as style, styleMap$1 as styleMap, styleTree$1 as styleTree };
