'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var Promise = require('bluebird');
var partition = require('lodash/partition');
var chalk = require('chalk');
var flatten = require('lodash/flatten');
var objectHash = require('object-hash');
var TreatError = require('../../dist/TreatError-e38bd3a1.browser.cjs.js');
var intersection = require('lodash/intersection');
var once = require('lodash/once');
var NodeTemplatePlugin = require('webpack/lib/node/NodeTemplatePlugin');
var NodeTargetPlugin = require('webpack/lib/node/NodeTargetPlugin');
var LibraryTemplatePlugin = require('webpack/lib/LibraryTemplatePlugin');
var SingleEntryPlugin = require('webpack/lib/SingleEntryPlugin');
var LimitChunkCountPlugin = require('webpack/lib/optimize/LimitChunkCountPlugin');
var ExternalsPlugin = require('webpack/lib/ExternalsPlugin');
var dedent = require('dedent');
var joi = require('@hapi/joi');
require('debug');

function _interopDefault (e) { return e && e.__esModule ? e : { 'default': e }; }

var Promise__default = /*#__PURE__*/_interopDefault(Promise);
var partition__default = /*#__PURE__*/_interopDefault(partition);
var chalk__default = /*#__PURE__*/_interopDefault(chalk);
var flatten__default = /*#__PURE__*/_interopDefault(flatten);
var objectHash__default = /*#__PURE__*/_interopDefault(objectHash);
var intersection__default = /*#__PURE__*/_interopDefault(intersection);
var once__default = /*#__PURE__*/_interopDefault(once);
var NodeTemplatePlugin__default = /*#__PURE__*/_interopDefault(NodeTemplatePlugin);
var NodeTargetPlugin__default = /*#__PURE__*/_interopDefault(NodeTargetPlugin);
var LibraryTemplatePlugin__default = /*#__PURE__*/_interopDefault(LibraryTemplatePlugin);
var SingleEntryPlugin__default = /*#__PURE__*/_interopDefault(SingleEntryPlugin);
var LimitChunkCountPlugin__default = /*#__PURE__*/_interopDefault(LimitChunkCountPlugin);
var ExternalsPlugin__default = /*#__PURE__*/_interopDefault(ExternalsPlugin);
var dedent__default = /*#__PURE__*/_interopDefault(dedent);
var joi__default = /*#__PURE__*/_interopDefault(joi);

const createContentHash = content => {
  return objectHash__default['default'].sha1(content).slice(0, 5).replace(/(^[0-9])/, '_$1');
};

var store = (() => {
  const treatModules = [];
  const themes = new Map();
  const themedCSSRequests = new Map();
  const cssResources = new Set();

  const getCSSResources = () => cssResources;

  const addCssRequests = (identifier, requests) => {
    requests.forEach(({
      resource
    }) => {
      cssResources.add(resource);
    });

    if (themes.size > 0) {
      themedCSSRequests.set(identifier, requests);
    }
  };

  const deleteCssRequests = identifier => {
    const requests = themedCSSRequests.get(identifier);
    requests.forEach(({
      resource
    }) => {
      cssResources.delete(resource);
    });
    themedCSSRequests.delete(identifier);
  };

  const getAllCssOwners = () => Array.from(themedCSSRequests.keys());

  const getAllThemedCssRequests = () => flatten__default['default'](Array.from(themedCSSRequests.values()));

  const getThemedCssModuleInfo = cssResource => {
    for (const [owner, requests] of themedCSSRequests.entries()) {
      for (const request of requests) {
        if (request.resource === cssResource) {
          return {
            owner,
            type: request.type,
            theme: request.theme
          };
        }
      }
    }

    throw new Error(`Couldn't find owner module for ${cssResource}`);
  };

  const addTheme = (theme, identifier, dependencies) => {
    themes.set(theme.themeRef, {
      theme,
      identifier,
      dependencies
    });
  };

  const getTheme = themeRef => themes.get(themeRef);

  const deleteModuleThemes = moduleIdentifier => {
    Array.from(themes.entries()).filter(([_, {
      identifier
    }]) => identifier === moduleIdentifier).forEach(([themeRef]) => themes.delete(themeRef));
  };

  const getThemesHash = () => createContentHash(Array.from(themes.entries()));

  const getThemes = () => Array.from(themes.values()).map(({
    theme
  }) => theme);

  const getThemeCount = () => themes.size;

  const getThemeDependencies = () => flatten__default['default'](Array.from(themes.values()).map(({
    dependencies
  }) => dependencies));

  const getThemeIdentifiers = () => Array.from(themes.values()).map(({
    identifier
  }) => identifier);

  const popAllModules = () => treatModules.splice(0, treatModules.length);

  const addModule = moduleIdentifier => treatModules.push({
    moduleIdentifier,
    themeHash: getThemesHash()
  });

  return {
    popAllModules,
    addModule,
    addTheme,
    getTheme,
    getCSSResources,
    getThemeCount,
    deleteModuleThemes,
    addCssRequests,
    deleteCssRequests,
    getAllThemedCssRequests,
    getAllCssOwners,
    getThemedCssModuleInfo,
    getThemesHash,
    getThemes,
    getThemeDependencies,
    getThemeIdentifiers
  };
});

TreatError.debug.formatters.r = ([currIndex, newPreIndex]) => currIndex > newPreIndex ? `${chalk__default['default'].green(currIndex)} -> ${chalk__default['default'].red(newPreIndex)}` : `${chalk__default['default'].red(currIndex)} -> ${chalk__default['default'].green(newPreIndex)}`;

const trace = TreatError.debug('treat:webpack-plugin:reindex');

const sortModules = (modules, {
  getPreIndex,
  getOwnerIndex,
  getThemeIndex
}) => {
  return modules.sort((a, b) => {
    if (a.type !== TreatError.THEMED && b.type === TreatError.THEMED) {
      return -1;
    }

    if (a.type === TreatError.THEMED && b.type !== TreatError.THEMED) {
      return 1;
    }

    if (a.type === TreatError.THEMED && b.type === TreatError.THEMED) {
      const themeIndexA = getThemeIndex(a);
      const themeIndexB = getThemeIndex(b);

      if (themeIndexA !== themeIndexB) {
        return themeIndexA - themeIndexB;
      }
    }

    const ownerIndexA = getOwnerIndex(a);
    const ownerIndexB = getOwnerIndex(b);

    if (ownerIndexA === ownerIndexB) {
      return getPreIndex(a) - getPreIndex(b);
    }

    return ownerIndexA - ownerIndexB;
  });
};

var reIndexModules = ((modules, {
  getPreIndex,
  getPostIndex,
  getOwnerIndex,
  getThemeIndex,
  setPreIndex,
  setPostIndex
}, target) => {
  trace('Sorting %s', target);
  const modulesToSort = modules.filter(m => {
    const hasOwnerIndex = typeof getOwnerIndex(m) === 'number';
    const shouldSort = hasOwnerIndex && (m.type !== TreatError.THEMED || typeof getThemeIndex(m) === 'number');

    if (!shouldSort) {
      trace('Ignoring %i from sorting. No %s index.', m.identifier, hasOwnerIndex ? 'theme' : 'owner');
    }

    return shouldSort;
  });
  const originalOrderModules = modulesToSort.slice().sort((a, b) => getPreIndex(a) - getPreIndex(b));
  const sortedModules = sortModules(modulesToSort, {
    getPreIndex,
    getOwnerIndex,
    getThemeIndex
  });
  originalOrderModules.map(module => {
    const swapModule = originalOrderModules[sortedModules.findIndex(m => m.identifier === module.identifier)];
    return [module, swapModule];
  }).filter(([a, b]) => a.identifier !== b.identifier).map(([moduleInfo, newModuleLocation]) => {
    return {
      moduleInfo,
      newPreIndex: getPreIndex(newModuleLocation),
      newPostIndex: getPostIndex(newModuleLocation)
    };
  }).forEach(({
    moduleInfo,
    newPreIndex,
    newPostIndex
  }) => {
    trace('Moving %r %i', [getPreIndex(moduleInfo), newPreIndex], moduleInfo.identifier);
    setPreIndex(moduleInfo, newPreIndex);
    setPostIndex(moduleInfo, newPostIndex);
  });
});

const trace$1 = TreatError.debug('treat:webpack-plugin:compiler');
const TWL = 'treat-webpack-loader';
const logMultiWebpackError = once__default['default'](() => {
  console.log(chalk__default['default'].red(dedent__default['default']`
  TreatWebpackPlugin: Error occured during treat file compilation.
  
  This error is usually caused by having multiple versions of webpack installed. 
  Ensure you only have a single version of webpack by running: '${chalk__default['default'].bold('yarn why webpack')}' or '${chalk__default['default'].bold('npm ls webpack')}'.
  `));
});
var makeTreatCompiler = (() => {
  const cache = new Map();

  const expireCache = changedFiles => {
    trace$1('Expire cache for files dependendent on: %O', changedFiles);
    Array.from(cache.entries()).forEach(([identifier, {
      fileDependencies
    }]) => {
      const expiredDependencies = intersection__default['default'](fileDependencies, changedFiles);

      if (expiredDependencies.length > 0) {
        trace$1('Expire cache for %i', identifier);
        cache.delete(identifier);
      }
    });
  };

  const getSource = async (loader, request) => {
    const identifier = loader._module.identifier();

    trace$1('Get compiled source: %i', identifier);
    const cachedValue = cache.get(identifier);

    if (cachedValue) {
      trace$1('Return cached source: %i', identifier);
      return cachedValue;
    }

    trace$1('No cached source. Compiling: %i', identifier);
    const compilationResult = await compileTreatSource(loader);
    cache.set(identifier, compilationResult);
    return compilationResult;
  };

  const getCompiledSource = async (loader, request) => {
    const {
      source,
      fileDependencies,
      contextDependencies
    } = await getSource(loader); // Set loader dependencies to dependecies of the child compiler

    fileDependencies.forEach(dep => {
      loader.addDependency(dep);
    });
    contextDependencies.forEach(dep => {
      loader.addContextDependency(dep);
    });
    return {
      source,
      dependencies: fileDependencies
    };
  };

  return {
    getCompiledSource,
    expireCache
  };
});

function getRootCompilation(loader) {
  var compiler = loader._compiler;
  var compilation = loader._compilation;

  while (compiler.parentCompilation) {
    compilation = compiler.parentCompilation;
    compiler = compilation.compiler;
  }

  return compilation;
}

function compileTreatSource(loader) {
  return new Promise__default['default']((resolve, reject) => {
    // Child compiler will compile treat files to be evaled during compilation
    const outputOptions = {
      filename: loader.resourcePath
    };
    const childCompiler = getRootCompilation(loader).createChildCompiler(TWL, outputOptions, [new NodeTemplatePlugin__default['default'](outputOptions), new LibraryTemplatePlugin__default['default'](null, 'commonjs2'), new NodeTargetPlugin__default['default'](), new SingleEntryPlugin__default['default'](loader.context, loader.resourcePath), new LimitChunkCountPlugin__default['default']({
      maxChunks: 1
    }), new ExternalsPlugin__default['default']('commonjs', 'treat')]);
    const subCache = 'subcache ' + __dirname + ' ' + loader.resourcePath;
    childCompiler.hooks.compilation.tap(TWL, compilation => {
      if (compilation.cache) {
        if (!compilation.cache[subCache]) compilation.cache[subCache] = {};
        compilation.cache = compilation.cache[subCache];
      }
    });
    let source;
    childCompiler.hooks.afterCompile.tapAsync(TWL, (compilation, callback) => {
      source = compilation.assets[loader.resourcePath] && compilation.assets[loader.resourcePath].source(); // Remove all chunk assets

      compilation.chunks.forEach(chunk => {
        chunk.files.forEach(file => {
          delete compilation.assets[file];
        });
      });
      callback();
    });

    try {
      childCompiler.runAsChild((err, entries, compilation) => {
        if (err) return reject(err);

        if (compilation.errors.length > 0) {
          return reject(compilation.errors[0]);
        }

        if (!source) {
          return reject(new Error("Didn't get a result from child compiler"));
        }

        resolve({
          source,
          fileDependencies: Array.from(compilation.fileDependencies),
          contextDependencies: Array.from(compilation.contextDependencies)
        });
      });
    } catch (e) {
      if (e.message.indexOf('No dependency factory available for this dependency type: SingleEntryDependency') > -1) {
        logMultiWebpackError();
      }

      reject(e);
    }
  });
}

const schema = joi__default['default'].object({
  test: joi__default['default'].any(),
  outputCSS: joi__default['default'].boolean(),
  outputLoaders: joi__default['default'].array().items(joi__default['default'].any()),
  localIdentName: joi__default['default'].string(),
  themeIdentName: [joi__default['default'].func(), joi__default['default'].string()],
  minify: joi__default['default'].boolean(),
  browsers: [joi__default['default'].string(), joi__default['default'].array().items(joi__default['default'].string())],
  hmr: joi__default['default'].boolean()
});

class ValidationError extends Error {
  constructor(error) {
    super();
    this.name = 'treat-webpack-plugin: Invalid options passed to treat plugin';
    this.message = `\n${error.annotate()}\n`;
    Error.captureStackTrace(this, this.constructor);
  }

}

var optionValidator = (options => {
  const {
    error
  } = schema.validate(options);

  if (error) {
    throw new ValidationError(error);
  }
});

const webpack4 = {
  isModuleUsed: (_compilation, module) => typeof module.used === 'boolean' ? module.used : true,
  getDependencyModule: (_compilation, dependency) => dependency.module,
  getModulePreOrderIndex: (_compilation, module) => module.index,
  setModulePreOrderIndex: (_compilation, module, index) => module.index = index,
  getModulePostOrderIndex: (_compilation, module) => module.index2,
  setModulePostOrderIndex: (_compilation, module, index) => module.index2 = index,
  getCGModulePreOrderIndex: (chunkGroup, module) => chunkGroup.getModuleIndex(module),
  setCGModulePreOrderIndex: (chunkGroup, module, index) => chunkGroup.setModuleIndex(module, index),
  getCGModulePostOrderIndex: (chunkGroup, module) => chunkGroup.getModuleIndex2(module),
  setCGModulePostOrderIndex: (chunkGroup, module, index) => chunkGroup.setModuleIndex2(module, index),
  isModuleInChunk: (_compilation, module, chunk) => chunk.containsModule(module),
  getModifiedFiles: watchCompiler => {
    // watchCompiler.watchFileSystem.watcher is undefined in some node environments.
    // Allow fallback to watchCompiler.watchFileSystem.wfs
    // https://github.com/s-panferov/awesome-typescript-loader/commit/c7da9ac82d105cdaf9b124ccc4c130648e59168a
    const watcher = watchCompiler.watchFileSystem.watcher || watchCompiler.watchFileSystem.wfs.watcher;
    return Object.keys(watcher.mtimes);
  },
  getModuleIssuer: (_compilation, module) => module.issuer
};
const webpack5 = {
  isModuleUsed: (compilation, module) => {
    const exportsInfo = compilation.moduleGraph.getExportsInfo(module);
    return exportsInfo.isModuleUsed('main');
  },
  getDependencyModule: (compilation, dependency) => compilation.moduleGraph.getModule(dependency),
  getModulePreOrderIndex: (compilation, module) => compilation.moduleGraph.getPreOrderIndex(module),
  setModulePreOrderIndex: (compilation, module, index) => compilation.moduleGraph.setPreOrderIndex(module, index),
  getModulePostOrderIndex: (compilation, module) => compilation.moduleGraph.getPostOrderIndex(module),
  setModulePostOrderIndex: (compilation, module, index) => compilation.moduleGraph.setPostOrderIndex(module, index),
  getCGModulePreOrderIndex: (chunkGroup, module) => chunkGroup.getModulePreOrderIndex(module),
  setCGModulePreOrderIndex: (chunkGroup, module, index) => chunkGroup.setModulePreOrderIndex(module, index),
  getCGModulePostOrderIndex: (chunkGroup, module) => chunkGroup.getModulePostOrderIndex(module),
  setCGModulePostOrderIndex: (chunkGroup, module, index) => chunkGroup.setModulePostOrderIndex(module, index),
  isModuleInChunk: (compilation, module, chunk) => compilation.chunkGraph.isModuleInChunk(module, chunk),
  getModifiedFiles: watchCompiler => watchCompiler.modifiedFiles ? Array.from(watchCompiler.modifiedFiles) : [],
  getModuleIssuer: (compilation, module) => compilation.moduleGraph.getIssuer(module)
};
var createCompat = (isWebpack5 => {
  if (isWebpack5) {
    return webpack5;
  }

  return webpack4;
});

const isProductionLikeMode = options => {
  return options.mode === 'production' || !options.mode;
};

const trace$2 = TreatError.debug('treat:webpack-plugin');
const TWP = 'treat-webpack-plugin';

const makeOptionDefaulter = prodLike => (option, {
  dev,
  prod
}) => {
  if (typeof option !== 'undefined') {
    return option;
  }

  return prodLike ? prod : dev;
};

class TreatPlugin {
  constructor(options = {}) {
    optionValidator(options);
    const {
      test = /\.treat\.(js|ts)$/,
      outputCSS = true,
      outputLoaders = ['style-loader'],
      localIdentName,
      themeIdentName,
      minify,
      browsers,
      hmr
    } = options;
    this.store = store();
    this.treatCompiler = makeTreatCompiler();
    this.test = test;
    this.minify = minify;
    this.localIdentName = localIdentName;
    this.themeIdentName = themeIdentName;
    this.loaderOptions = {
      outputCSS,
      outputLoaders,
      browsers,
      hmr
    };
  }

  apply(compiler) {
    const isWebpack5 = Boolean(compiler.webpack && compiler.webpack.version);
    const compat = createCompat(isWebpack5);
    compiler.hooks.watchRun.tap(TWP, watchCompiler => {
      const modifiedFiles = compat.getModifiedFiles(watchCompiler);
      this.treatCompiler.expireCache(modifiedFiles);
    });
    compiler.hooks.thisCompilation.tap(TWP, compilation => {
      let allCssModules = [];
      let usedCssModules;

      const rebuildModule = module => new Promise__default['default'](resolve => {
        compilation.rebuildModule(module, () => {
          resolve();
        });
      });

      compilation.hooks.finishModules.tapPromise(TWP, async () => {
        // Filter out css modules that no longer exist (watch mode)
        this.store.getAllCssOwners().filter(identifier => !Boolean(compilation.findModule(identifier))).forEach(identifier => {
          this.store.deleteCssRequests(identifier);
        }); // Filter out theme treat modules that no longer exist (watch mode)

        this.store.getThemeIdentifiers().filter(identifier => !Boolean(compilation.findModule(identifier))).forEach(identifier => {
          this.store.deleteModuleThemes(identifier);
        });
        const finalThemeHash = this.store.getThemesHash();
        const builtTreatModules = this.store.popAllModules().filter(({
          moduleIdentifier
        }) => !this.store.getThemeIdentifiers().includes(moduleIdentifier)); // Rebuild all non theme treat modules that built before all themes were registered

        const staleModules = builtTreatModules.filter(({
          themeHash
        }) => themeHash != finalThemeHash);
        await Promise__default['default'].map(staleModules, async ({
          moduleIdentifier
        }) => {
          trace$2('Rebuiling stale module: %i', moduleIdentifier);
          const treatModule = compilation.findModule(moduleIdentifier);

          if (!treatModule) {
            // In cases where the build fails, sometimes modules are not present
            // on the compilation object. Skip rebuilding in these cases.
            trace$2("Bail on rebuild of stale module: %i Module doesn't exist on compilation", moduleIdentifier);
            return;
          }

          await rebuildModule(treatModule);
        }); // Modules should be upto date so clear rebuild queue

        this.store.popAllModules();

        if (builtTreatModules.length > 0) {
          // Rebuild all theme modules to ensure they contain upto date css
          await Promise__default['default'].map(this.store.getThemeIdentifiers(), async themeModuleIdentifier => {
            trace$2('Rebuiling theme module: %i', themeModuleIdentifier);
            const treatModule = compilation.findModule(themeModuleIdentifier);

            if (!treatModule) {
              // In cases where the build fails, sometimes modules are not present
              // on the compilation object. Skip rebuilding in these cases.
              trace$2("Bail on rebuild of theme module: %i Module doesn't exist on compilation", themeModuleIdentifier);
              return;
            }

            await rebuildModule(treatModule);
          });
        }

        const cssResources = this.store.getAllThemedCssRequests().map(({
          resource
        }) => resource);

        const getTreatCssResource = module => {
          if (module.matchResource && cssResources.includes(module.matchResource)) {
            return module.matchResource;
          }

          if (module.type === 'css/mini-extract') {
            const issuer = compat.getModuleIssuer(compilation, module);
            return getTreatCssResource(issuer);
          }

          return null;
        };

        for (const module of compilation._modules.values()) {
          const treatCssResource = getTreatCssResource(module);

          if (treatCssResource) {
            const {
              owner,
              type,
              theme
            } = this.store.getThemedCssModuleInfo(treatCssResource);
            const themeIdentifier = theme ? this.store.getTheme(theme).identifier : null;
            allCssModules.push({
              type,
              module: compilation.findModule(module.identifier()),
              owner: compilation.findModule(owner),
              themeModule: theme ? compilation.findModule(themeIdentifier) : null,
              identifier: module.identifier()
            });
          }
        }
      });
      compilation.hooks.optimizeDependencies.tap( // Running during advanced stage means export usage info has been added to the graph already
      {
        name: TWP,
        stage: 10
      }, () => {
        // Some modules may not be used and their css can be safely be removed from the chunks
        const cssModuleGroups = partition__default['default'](allCssModules, ({
          module,
          owner,
          themeModule
        }) => {
          const ownerIsUsed = compat.isModuleUsed(compilation, owner);
          const themeIsUsed = !themeModule || compat.isModuleUsed(compilation, themeModule);
          const cssModuleIsUsed = ownerIsUsed && themeIsUsed;

          if (!cssModuleIsUsed) {
            trace$2('CSS Module marked for removal due to unused %s %m', chalk__default['default'].yellow(ownerIsUsed ? 'theme' : 'owner'), module);
          }

          return cssModuleIsUsed;
        });
        usedCssModules = cssModuleGroups[0];
        const cssModulesToRemove = cssModuleGroups[1];

        if (cssModulesToRemove.length > 0) {
          const themeModules = this.store.getThemeIdentifiers().map(ident => compilation.findModule(ident));
          const cssModulesIdentsToRemove = cssModulesToRemove.map(({
            identifier
          }) => identifier);

          for (const themeModule of themeModules) {
            const depsToRemove = themeModule.dependencies.filter(dep => {
              const depModule = compat.getDependencyModule(compilation, dep);
              const shouldRemove = depModule && cssModulesIdentsToRemove.includes(depModule.identifier());

              if (shouldRemove) {
                trace$2('Mark %m for clean-up in %m', depModule, themeModule);
              }

              return shouldRemove;
            });

            for (const dep of depsToRemove) {
              themeModule.removeDependency(dep);
            }

            compilation.removeReasonsOfDependencyBlock(themeModule, {
              dependencies: depsToRemove,
              blocks: []
            });
          }
        }
      });
      compilation.hooks.afterChunks.tap(TWP, chunks => {
        try {
          // afterChunks hook means module/chunk order properties have been set
          // We can now correct those values by referencing the ordering of the owner treat file
          for (const chunk of chunks) {
            const cssModulesInChunk = usedCssModules.filter(({
              module
            }) => compat.isModuleInChunk(compilation, module, chunk));

            for (const chunkGroup of chunk.groupsIterable) {
              // Corrects ChunkGroup._moduleIndicies
              // Used for mini-extract-css-plugin ordering
              reIndexModules(cssModulesInChunk, {
                getPreIndex: ({
                  module
                }) => compat.getCGModulePreOrderIndex(chunkGroup, module),
                getPostIndex: ({
                  module
                }) => compat.getCGModulePostOrderIndex(chunkGroup, module),
                getOwnerIndex: ({
                  owner
                }) => compat.getModulePreOrderIndex(compilation, owner),
                getThemeIndex: ({
                  themeModule
                }) => compat.getModulePreOrderIndex(compilation, themeModule),
                setPreIndex: ({
                  module
                }, i) => compat.setCGModulePreOrderIndex(chunkGroup, module, i),
                setPostIndex: ({
                  module
                }, i) => compat.setCGModulePostOrderIndex(chunkGroup, module, i)
              }, 'ChunkGroup moduleIndicies');
            }
          } // Corrects Module.index/index2


          reIndexModules(usedCssModules, {
            getPreIndex: ({
              module
            }) => compat.getModulePreOrderIndex(compilation, module),
            getPostIndex: ({
              module
            }) => compat.getModulePostOrderIndex(compilation, module),
            getOwnerIndex: ({
              owner
            }) => compat.getModulePreOrderIndex(compilation, owner),
            getThemeIndex: ({
              themeModule
            }) => compat.getModulePreOrderIndex(compilation, themeModule),
            setPreIndex: ({
              module
            }, i) => compat.setModulePreOrderIndex(compilation, module, i),
            setPostIndex: ({
              module
            }, i) => compat.setModulePostOrderIndex(compilation, module, i)
          }, 'Module.index/index2');
          this.store.getThemeIdentifiers().map(themeModuleIdentifier => compilation.findModule(themeModuleIdentifier)).forEach(themeModule => {
            const {
              dependencies
            } = themeModule;
            const relevantDependencies = usedCssModules.map(moduleInfo => ({ ...moduleInfo,
              dependency: dependencies.find(d => {
                const module = compat.getDependencyModule(compilation, d);
                return module && module.identifier() === moduleInfo.identifier;
              })
            })).filter(({
              dependency
            }) => dependency); // Corrects Dependency.sourceOrder
            // Used for regular import chunk sorting, e.g. style-loader

            reIndexModules(relevantDependencies, {
              getPreIndex: ({
                dependency
              }) => dependency.sourceOrder,
              getOwnerIndex: ({
                owner
              }) => compat.getModulePreOrderIndex(compilation, owner),
              getThemeIndex: ({
                themeModule
              }) => compat.getModulePreOrderIndex(compilation, themeModule),
              setPreIndex: ({
                dependency
              }, i) => {
                dependency.sourceOrder = i;
              },
              // post index not required for dependencies
              getPostIndex: () => {},
              setPostIndex: () => {}
            }, 'Dependency.sourceOrder');
          });
        } catch (e) {
          compilation.errors.push(new TreatError.TreatError(e));
        }
      });
    });
    compiler.hooks.normalModuleFactory.tap(TWP, nmf => {
      nmf.hooks.afterResolve.tap(TWP, result => {
        if (this.store.getCSSResources().has(result.matchResource)) {
          result.settings = Object.assign({}, result.settings, {
            sideEffects: true
          });
        }
      });
    });
    const optionDefaulter = makeOptionDefaulter(isProductionLikeMode(compiler.options));
    compiler.options.module.rules.splice(0, 0, {
      test: this.test,
      use: [{
        loader: require.resolve('treat/webpack-plugin/loader'),
        options: { ...this.loaderOptions,
          minify: optionDefaulter(this.minify, {
            dev: false,
            prod: true
          }),
          localIdentName: optionDefaulter(this.localIdentName, {
            dev: '[name]-[local]_[hash:base64:5]',
            prod: '[hash:base64:5]'
          }),
          themeIdentName: optionDefaulter(this.themeIdentName, {
            dev: '_[name]-[local]_[hash:base64:4]',
            prod: '[hash:base64:4]'
          }),
          store: this.store,
          treatCompiler: this.treatCompiler
        }
      }]
    }, {
      test: /\.treatcss$/,
      sideEffects: true,
      use: [...this.loaderOptions.outputLoaders, {
        loader: 'css-loader',
        options: {
          modules: false,
          url: false
        }
      }]
    });
  }

}

exports.TreatPlugin = TreatPlugin;
